/*
 *
 * misc.c -- miscellaneous routines
 *
 */

#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <time.h>

#include "scram.h"
#include "globals.h"

FILE *gcgptr;
FILE *evalptr;
FILE *logptr;

char saved_rack[8];

char spaces[81] =
"                                        \
                                        ";

void read_words() {
  FILE *fp;
  char inp[16], tinp[16];
  int i, tlen, rlen;
  long pt, wr;
	
  fp = fopen( DICT_FILE, "r" );
  if( fp == NULL ) {
    move( 4, 0 );
    printw( "Error: could not find dictionary file '%s'\n", DICT_FILE );
    refresh();
    sleep(2);
    exit_window();
    exit( 1 );
  }
  printw( "Reading dictionary...\n" );
  refresh();
  pt = 0;
  wr = 0;
  tlen = 2;
  wlen[tlen] = wr;
  tinp[0] = '\0';
  /* Errors in the dictionary file lead to an immediate exit without
     bothering with the log file or anything else, as the error
     prevents the program from running at all (ever) until fixed. */
  while( fscanf( fp, "%s", inp ) != EOF ) {
    rlen = strlen( inp );
    if( rlen < 2 || rlen > 15 ) {
      move( 4, 0 );
      printw( "Error in dictionary file at word %s: words must be between 2 and 15\n", inp );
      printw( "characters long.\n" );
      exit_window();
      exit( 1 );
    }
    if( rlen < strlen( tinp ) ||
	( strcmp( tinp, inp ) > 0 && strlen( tinp ) == rlen ) ) {
      move( 4, 0 );
      printw( "Error in dictionary file at the words %s and %s.\n",
	      tinp, inp );
      printw( "The dictionary must be sorted from shortest words to longest, and\n" );
      printw( "alphabetically within each list of words of the same length." );
      exit_window();
      exit( 1 );
    }
    for( i = 0; i < rlen; i++ ) {
      if( inp[i] >= 'a' && inp[i] <= 'z' ) inp[i] = inp[i] - 'a' + 'A';
      else if( inp[i] < 'A' || inp[i] > 'Z' ) {
	move( 4, 0 );
	printw( "Error in dictionary file at word %s: words must contain alphabetic characters\n", inp );
	printw( "only.\n" );
	exit_window();
	exit( 1 );
      }
    }
    if( rlen > tlen ) {
      for( i = tlen + 1; i <= rlen; i++ ) wlen[i] = wr;
      tlen = rlen;
    }
    strcpy( &words[pt], inp );
    wptr[wr++] = pt;
    pt += rlen + 1;
    strcpy( tinp, inp );
  }
  words[pt] = '\0';
  wptr[wr] = pt;
  dict_size = pt;
  for( i = tlen + 1; i < 17; i++ ) wlen[i] = wr;
  fclose( fp );
}

void clear_rect( y1, x1, y2, x2 )
  int y1, x1, y2, x2;
{
  int i;

  for( i = y1; i <= y2; i++ ) {
    move( i, x1 );
    printw( "%s", &spaces[79 - ( x2 - x1 )] );
  }
  move( y1, x1 );
}

void clear_prompt() {
  clear_rect( 19, 0, 23, 78 );
  move( 19, 0 );
}

void clear_turn() {
  clear_rect( 18, 0, 18, 79 );
  move( 18, 0 );
}

void press_return() {
  char c;
  char cn;
  char i, abort, srstatus;
  char note[80];

  move( 23, 0 );
  printw( "Press <RETURN> -> " );
  refresh();
  do {
    c = getch();
    if (c != '\n')
      printw("%c", c);
    if( c == 'q' || c == 'Q' ) {
      game_abort = 1;
      /* Save. */
    } else if ( c == 's' || c == 'S' ) {
      clear_rect( 23, 0, 22, 79 );
      move (23, 0);
      if (sr_status == 0) {
	srstatus=save_game();
      } else {
	clear_rect( 23, 0, 22, 79 );
	move (23, 0);
	printw("Cannot save right now");
	refresh();
	sleep(2);
      }
      /* Fake exit. */
      c='\n';
    } else if ( c == 'r' || c == 'R') {
      /* Ask to save active game first? No. Deliberate
	 user decision here.*/
      if (sr_status == 0) {
	srstatus=restore_game();
	/* Indicate restored game. Nothing to change on failure. */
	if (srstatus==0) {
	  sr_status = 1;
	}
      } else {
	clear_rect( 23, 0, 22, 79 );
	move (23, 0);
	printw("Cannot restore right now");
	refresh();
	sleep(2);
      }
      /* Fake to exit loop. */
      c = '\n';
      /* Start on user notes */
      /* Backspace, C-h don't work */
    } else if ( c == 'n' || c == 'N' ) {
      clear_rect ( 23, 1, 23, 79 );
      move ( 23, 0 );
      printw( "Brief note: ");
      i=0;
      abort=0;
      do { 
	note[i]='\0';
	cn=getch();
	switch (cn) {
	case '\n' : break;
	  /* backsp */
	case 8 :
	case 127 :
	  if (i > 0){
	    i--;
	    note[i]='\0';
	    /* need to backspace over prev char and clear */
	    printw("\b \b");
	  }
	  break;
	case 27 :
	  abort = 1;
	default : 
	  note[i]=cn;
	  i++;
	  printw("%c",cn);
	  break;
	}
      } while (cn != '\n' && i < 67 && abort==0);
      clear_rect ( 23, 1, 23, 79 );
      if ((abort == 0) && (strlen(note) > 1))
	fprintf(gcgptr, "#note: %s\n", note);
      c='\n';
    }
  } while( c != '\n' );
}

void print_board() {
  int i, j;

  clear_rect( BOARD_UY, BOARD_LX, BOARD_LY, BOARD_RX );
  for( i = 1; i < 16; i++ ) {
    for( j = 1; j < 16; j++ ) {
      move( BOARD_UY + i - 1, BOARD_LX + j * 2 - 2 );
      if( board[i][j] >= 'A' ) standout();
      if( ( i == b1y && j == b1x ) || ( i == b2y && j == b2x ) )
	addch( board[i][j] - 'A' + 'a' );
      else if( board[i][j] == CH_DW ) addch( 'o' );
      else addch( board[i][j] );
      if( board[i][j] >= 'A' ) standend();
    }
  }
  refresh();
}

void write_board() {
  /* Output analysis board to eval file. */
  int i, j, index;
  char row[34];
	
  /* Print column (x) notation. */
  fprintf(evalptr, "\n    A B C D E F G H I J K L M N O\n");
  for( i = 1; i < 16; i++ ) {
    strcpy(row,"                                ");
    /* Show row (y) notation. */
    fprintf(evalptr, "%3d ", i); 
    for( j = 1; j < 16; j++ ) {
      index=j*2-2;
      row[index] = board[i][j];
      if( ( i == b1y && j == b1x ) || ( i == b2y && j == b2x ) )
	row[index] = row[index] - 'A' + 'a';
      else if(row[index] == CH_DW ) row[index] = 'o';
    }
    row[33] = '\0';
    fprintf(evalptr, "%s\n", row);
  }
}

void print_tiles() {
  int i, j;

  clear_rect( TILE_RACKS_Y, TILE_RACKS_X,
	      TILE_RACKS_Y + 7, TILE_RACKS_X + 19 );
  for( i = 0; i < players; i++ ) {
    move( TILE_RACKS_Y + i * 3, TILE_RACKS_X - 4 );
    if( plr_skills[i] != 0 ) printw( "(%d)", plr_skills[i] );
    move( TILE_RACKS_Y + i * 3, TILE_RACKS_X + 1 );
    if( i == human_player )
      printw( " You:" );
    else
      printw( "%s:", you[i] );
    for( j = 0; j < 7; j++ ) {
      addch( ' ' );
      if( plr_tiles[i][j] > 0 ) {
	if( i == human_player || game_done == players )
	  addch( plr_tiles[i][j] );
	else addch( '-' );
      } else addch( ' ' );
    }
    move( TILE_RACKS_Y + i * 3 + 1, TILE_RACKS_X );
    printw( "Score: %3d", plr_scores[i] );
  }
  move( TILE_RACKS_Y + 12, TILE_RACKS_X - 2 );
  i = 0;
  if( b1x != 0 ) i++;
  if( b2x != 0 ) i++;
  printw( "Blank tiles placed: %d", i );
  if( b1x != 0 ) {
    move( TILE_RACKS_Y + 13, TILE_RACKS_X - 1 );
    printw( "Row %2d, Column %2d (%c)", b1y, b1x, board[b1y][b1x] );
  }
  if( b2x != 0 ) {
    move( TILE_RACKS_Y + 14, TILE_RACKS_X - 1 );
    printw( "Row %2d, Column %2d (%c)", b2y, b2x, board[b2y][b2x] );
  }
  move( TILE_RACKS_Y + human_player * 3, TILE_RACKS_X - 1 );
  refresh();
}

char draw_tile() {
  int i, tleft;
  int tile;

  tleft = 0;
  for( i = 0; i < 27; i++ )
    tleft += tiles_left[i];

  if( tleft == 0 ) return( 0 );
  tile = get_random() % tleft;
  for( i = 0;; i++ ) {
    if( tile - tiles_left[i] >= 0 )
      tile -= tiles_left[i];
    else
      break;
  }
  tiles_left[i]--;
  print_tiles_left();
  return( i == 26 ? CH_BL : i + 'A' );
}

void print_tiles_left() {
  int i, tleft;

  tleft = 0;
  for( i = 0; i < 27; i++ )
    tleft += tiles_left[i];

  move( TILES_LEFT_Y, TILES_LEFT_X );
  printw( "Tiles left to draw: %3d", tleft );
}

void count_letters () {

  int plr, tileno, i;

  /* Count tiles in pool. Easy. */
  for ( i=0; i < 27; i++ )
    un_tiles[i] = tiles_left[i];

  /* Add in tiles for computer players. A little harder. */
  for ( plr = 0; plr < players; plr++) {
    if (plr != human_player) {
      for (i=0; i<7; i++) {
	/* Don't include missing rack spaces! */
	if (plr_tiles[plr][i] != 0) {
	  /* Asterisk? */
	  if (plr_tiles[plr][i] == '*') {
	    un_tiles[26]++;
	  } else {
	    un_tiles[ (plr_tiles[plr][i] - 'A') ]++;
	  }
	}
      }
    }
  }
}

void print_passed_turns() {
  if( game_done == 0 ) clear_rect( 1, 50, 1, 79 );
  else {
    move( 1, 63 );
    printw( "Passed turns:  %d", game_done );
  }
  refresh();
}

void print_whos_move( plr )
     int plr;
{
  clear_turn();
  printw( "*** %s turn to play  ", your[plr] );
}

void print_version() {
  clear();
  printw( "*** %s ***\n\n", version );
}

void screen_redraw() {
  print_version();
  print_info();
  print_un_tiles();
  print_board();
  print_tiles_left();
  print_passed_turns();
  print_tiles();
  print_whos_move( human_player );
}

char get_key() {
  char c;

  c = getch();
  if( c >= 'a' && c <= 'z' ) c = c - 'a' + 'A';
  return( c );
}

void seed_random() {
  long rightnow, time();
  time_t t;
  rightnow = time( 0 );
  srandom((unsigned) time(&t));
}

long get_random() {
  return( (long) random() );
}

void open_eval () {
  int i, stop;
  char fname[20];
  
  i = 1;
  stop = 0;
  strcpy(fname, "001.eval.txt");
  do {
    if( access( fname, F_OK ) == 0 ) {
      i++;
      sprintf(fname, "%03d.eval.txt", i);
    } else {
      evalptr = fopen(fname, "w+");
      stop = 1;
    }
  } while (stop==0);
  strcpy(evalname, fname);

}
void open_log () {

  logptr=fopen(logname, "a+");
  if (logptr == NULL) {
    printf("Can't open Scrambler log file ... aborting\n");
    exit (1);
  }
}

void open_gcg () {
  int i, stop;
  char fname[20];
  
  i = 1;
  stop = 0;
  strcpy(fname, "001.gcg");
  do {
    if( access( fname, F_OK ) == 0 ) {
      i++;
      sprintf(fname, "%03d.gcg", i);
    } else {
      gcgptr = fopen(fname, "w+");
      stop = 1;
    }
  } while (stop==0);
  strcpy(gcgname, fname);
}

void gcg_rack (int plr) {
  int i;

  for (i=0;i<7;i++)
    saved_rack[i]='\0';

  for ( i = 0; i < 7; i++ ) {
    if (plr_tiles[plr][i] != '\0') {
      if (plr_tiles[plr][i] == '*') {
	fprintf(gcgptr, "?");
	saved_rack[i] = '?';
      } else {
	fprintf(gcgptr, "%c", plr_tiles[plr][i]);
	saved_rack[i] = plr_tiles[plr][i];
      }
    }
  }
}

void eval_rack (int plr) {
  int i;

  for ( i = 0; i < 7; i++ ) {
    if (plr_tiles[plr][i] != '\0') {
      if (plr_tiles[plr][i] == '*') {
	fprintf(evalptr, "?");
	saved_rack[i] = '?';
      } else {
	fprintf(evalptr, "%c", plr_tiles[plr][i]);
	saved_rack[i] = plr_tiles[plr][i];
      }
    }
  }
}

void check_dick () {

  FILE *fp;
  int p1;
  char c;

  /* Inhibit s/r on next press_return. */
  sr_status = -1;
  press_return();
  clear_prompt();
  printw( "  The computer's dictionary was modified this game.\n" );
  printw( "  Would you like to save the new dictionary (y/n)? " );
  refresh();
  do {
    c = get_key();
  } while( c != 'Y' && c != 'y' && c != 'n' && c != 'N' );
  if( c == 'Y' || c == 'y') {
    printw( "\n\n  Writing new dictionary..." );
    refresh();
    fp = fopen( DICT_FILE, "w+" );
    if( fp == NULL ) {
      printw( "error writing new dictionary!\n" );
      refresh();
    } else {
      for( p1 = 0; p1 < wlen[16]; p1++ )
	fprintf( fp, "%s\n", &(words[wptr[p1]]) );
      fclose( fp );
    }
  }
}

void aborthandler (int signo) {
  fprintf(gcgptr, "\n#note system/user abort no. %d\n", signo);
  fprintf(evalptr, "\n\n System/User abort\n\n");
  clear_rect ( 23, 1, 23, 79 );
  move ( 23, 0 );
  printw("Caught signal %d.\n", signo);
  /* On SIGSEGV or SIGABRT try to save. It may be hopeless and the
     save may be bad, but it might be of use in debugging the
     occasional segfaults seen during the computer move. */
  if (signo == 11 || signo == 6) {
    printw(" ... will attempt to save game for debugging.\n");
    printw(" Please send %s, %s, and \n", gcgname, evalname);
    printw(" save file to: scrambler-debug@bobnewell.net\n");
    printw(" Mahalo for your help in improving Scrambler.");
    refresh();
    sleep(3);
    save_game();
  }
  /* Encode signal number and pass to termination routine. */
  game_abort=signo+10;
  thatsallfolks();
}

void thatsallfolks() {

  time_t now;
  
  if (gcgptr != NULL) fclose(gcgptr);
  if (evalptr != NULL) fclose(evalptr);

  /* Offer to keep dictionary changes. This is risky when there has
     been an error condition. Backups recommended. */
  if( dict_changed == 1 ) check_dick();
  exit_window();

  /* 0 normal exit
     1 user quit
     2 eval_score logic error
     3 eval_score out of memory (compmove)
     4 eval_score out of memory (scorent)
     5 display_words logic error
     6 enter_name logic error
     Anything else is 10+signal number. */
  if (logptr != NULL) {
    time(&now);
    if (game_abort == 1) {
      fprintf(logptr, "USER QUIT   %s %s %s", gcgname, evalname, ctime(&now));
    } else if (game_abort > 10) {
      fprintf(logptr, "SYS ABORT   %d %s %s %s",
	      game_abort-10, gcgname, evalname, ctime(&now));
    } else if (game_abort == 0) {
      fprintf(logptr, "GAME END    %s %s %s", gcgname, evalname, ctime(&now));
    } else if (game_abort > 0) {
      fprintf(logptr, "INT. ERROR  %d %s %s %s", game_abort, gcgname, evalname, ctime(&now));
    } else {
      fprintf(logptr, "UNKNOWN END %s %s %s", gcgname, evalname, ctime(&now));
    }  
    fclose(logptr);
  }
  exit(0);
}


