/*
 *
 * smain.c -- main program
 *
 */

#include <time.h>
#include <signal.h>
#include <unistd.h>

#include "scram.h"
#include "globals.h"

void set_up_window() {
  WINDOW *screen;

  screen = initscr();
  refresh();
  /* Screen size check reinstated. Funny things can otherwise
     happen. */
  if( screen->_maxx < 79 || screen->_maxy < 23 ) {
    endwin();
    fprintf( stderr, "Need at least a 24x80 screen\n" );
    exit( 1 );
  }
  cbreak();
  noecho();
}

void exit_window() {
  char c;

  move( 23, 0 );
  printw( " Press any key to exit. Game records are in\n " );
  printw( "   %s and %s\n", gcgname, evalname);
  refresh();
  c = getch();
  endwin();
}

void init_board() {
  int i, j;

  for( i = 1; i < 16; i++ )
    for( j = 1; j < 16; j++ )
      board[i][j] = CH_EM;
  board[1][1] = CH_TW;
  board[8][1] = CH_TW;
  board[8][8] = CH_DW;
  for( i = 2; i < 6; i++ ) {
    board[i][i] = CH_DW;
  }
  board[2][6] = CH_TL;
  board[6][2] = CH_TL;
  board[6][6] = CH_TL;
  board[1][4] = CH_DL;
  board[4][1] = CH_DL;
  board[3][7] = CH_DL;
  board[7][3] = CH_DL;
  board[8][4] = CH_DL;
  board[7][7] = CH_DL;
  for( i = 1; i < 9; i++ )
    for( j = 1; j < 8; j++ ) {
      board[16-j][i] = board[i][j];
      board[16-i][16-j] = board[i][j];
      board[j][16-i] = board[i][j];
    }
}

void print_info() {
  /* This now only prints board symbols. Tiles/values have been
     separated out for the tiles remaining feature. */
  move( INFO_Y    , INFO_X );
  printw( "Special symbols:" );
  move( INFO_Y + 1, INFO_X + 2 );
  printw( "%c: Double letter score", CH_DL );
  move( INFO_Y + 2, INFO_X + 2 );
  printw( "%c: Triple letter score", CH_TL );
  move( INFO_Y + 3, INFO_X + 2 );
  printw( "%c: Double word score", 'o' );
  move( INFO_Y + 4, INFO_X + 2 );
  printw( "%c: Triple word score", CH_TW );
  move( INFO_Y + 5, INFO_X + 2 );
  printw( "%c: Blank tile", CH_BL );	
}

void print_un_tiles () {
  count_letters();
  move( INFO_Y + 7, INFO_X );
  printw( "Tiles unseen & values:" );
  move( INFO_Y + 8, INFO_X + 2 );
  printw( "%dA1  %dH4  %dO1  %dV4",
	  un_tiles[0], un_tiles[7], un_tiles[14], un_tiles[21] );
  move( INFO_Y + 9, INFO_X + 2 );
  printw( "%dB3  %dI1  %dP3  %dW4",
	  un_tiles[1], un_tiles[8], un_tiles[15], un_tiles[22] );
  move( INFO_Y +10, INFO_X + 2 );
  printw( "%dC3  %dJ8  %dQ10 %dX8",
	  un_tiles[2], un_tiles[9], un_tiles[16], un_tiles[23] );
  move( INFO_Y +11, INFO_X + 2 );
  printw( "%dD2  %dK5  %dR1  %dY4",
	  un_tiles[3], un_tiles[10], un_tiles[17], un_tiles[24] );
  move( INFO_Y +12, INFO_X + 1 );
  /* Deal with letter E starting with >9 tiles and causing formatting
     issues when it drops from double digits to single digit. This may
     look strange, but it works. */
  if (un_tiles[4] > 9) {
    printw( "%dE1  %dL1  %dS1  %dZ10",
	    un_tiles[4], un_tiles[11], un_tiles[18], un_tiles[25] );
  } else {
    printw( " %dE1  %dL1  %dS1  %dZ10 ",
	    un_tiles[4], un_tiles[11], un_tiles[18], un_tiles[25] );
  }
  move( INFO_Y +13, INFO_X + 2 );
  printw( "%dF4  %dM3  %dT1  %d*0",
	  un_tiles[5], un_tiles[12], un_tiles[19], un_tiles[26] );
  move( INFO_Y +14, INFO_X + 2 );
  printw( "%dG2  %dN1  %dU1",
	  un_tiles[6], un_tiles[13], un_tiles[20] );
  refresh();
}

void print_first_tiles () {
  move( INFO_Y + 7, INFO_X );
  printw( "Tile distribs & values:" );
  move( INFO_Y + 8, INFO_X + 2 );
  printw( "9A1  2H4  8O1  2V4" );
  move( INFO_Y + 9, INFO_X + 2 );
  printw( "2B3  9I1  2P3  2W4" );
  move( INFO_Y +10, INFO_X + 2 );
  printw( "2C3  1J8  1Q10 1X8" );
  move( INFO_Y +11, INFO_X + 2 );
  printw( "4D2  1K5  6R1  2Y4" );
  move( INFO_Y +12, INFO_X + 1 );
  printw( "12E1  4L1  4S1  1Z10" );
  move( INFO_Y +13, INFO_X + 2 );
  printw( "2F4  2M3  6T1  2%c0", CH_BL );
  move( INFO_Y +14, INFO_X + 2 );
  printw( "3G2  6N1  4U1" );
  refresh();
}

void init_tiles() {
  int i, j;

  for( i = 0; i < 27; i++ )
    tiles_left[i] = letters[i].tiles;
  for( i = 0; i < 4; i++ )
    for( j = 0; j < 7; j++ )
      plr_tiles[i][j] = 0;
  b1x = 0;
  b2x = 0;
  b1y = 0;
  b2y = 0;
  normal_sum = 0;
  for( i = 0; i < 40; i++	) normal_sum += normal_dist[i];
}

void get_num_players() {
  char c;
  int i, j, srstatus;
  char ft[4], cs[7];

  for( ;; ) {
    clear_prompt();
    printw( "  Number of computer players (1-3) or 'r' to restore saved game: " );
    refresh();
    c = getch();
    if (c == 'r' || c == 'R') {
      clear_prompt();
      srstatus=restore_game();
      if (srstatus==0) {
	sr_status = 1;
	clear_rect( 23, 0, 22, 79 );
	move (23, 0);
	printw("Restored.");	
	refresh();
	sleep(1);
      }
      if (sr_status == 1) break;
    } else if ( c >= '1' && c <= '3' ){
      break; 
    }
  }

  if (sr_status == -1) {
    players = c - '0' + 1;
    for( ;; ) {
      init_tiles();
      for( i = 0; i < players; i++ ) ft[i] = draw_tile();
      for( i = 1; i < players; i++ )
	if( ft[0] == ft[i] ) break;
      if( i != players ) continue;
      move( 20, 0 );
      printw( "  All players draw a tile.  You: %c", ft[0] );
      for( i = 1; i < players; i++ ) {
	printw( "  Player %d: %c", i + 1, ft[i] );
	if( ft[i] == CH_BL ) ft[i] = 'A' - 1;
      }
      j = players;
      for( i = 1; i < players; i++ )
	if( ft[0] < ft[i] ) j--;
      if( j == 1 ) strcpy( cs, "first" );
      if( j == 2 ) strcpy( cs, "second" );
      if( j == 3 ) strcpy( cs, "third" );
      if( j == 4 ) strcpy( cs, "fourth" );
      printw( "\n  You will play %s.", cs );
      human_player = j - 1;
      print_tiles_left();
      sr_status = -1;
      press_return();
      clear_prompt();
      refresh();
      break;
    }

    init_tiles();
    j = '1';
    for( i = 0; i < 4; i++ ) {
      if( i == human_player ) {
	strcpy( &(your[i][0]), "Your" );
	strcpy( &(you[i][0]), "You" );
      } else {
	strcpy( &(your[i][0]), "CPU 's" );
	strcpy( &(you[i][0]), "CPU " );
	your[i][3] = j;
	you[i][3] = j++;
      }
      plr_skills[i] = 0;
    }

    for( i = 0; i < players; i++ ) {
      print_tiles();
      if( i == human_player ) continue;
      clear_prompt();
      printw( "  Select skill level 1-%d for %s, where 1 is beginner, %d is advanced, and\n", SKILL_EXPERT, you[i], SKILL_LEVELS );
      printw( "  %d is expert. -> ", SKILL_EXPERT );
      refresh();
      for( ;; ) {
	c = getch();
	/* The hard-wired max is really not good. */
	if( c >= '1' && c <= '9' ) break;
      }
      plr_skills[i] = c - '0';
    }
	
    for( i = 0; i < players; i++ ) {
      plr_scores[i] = 0;
      for( j = 0; j < 7; j++ )
	plr_tiles[i][j] = draw_tile();
    }
    print_tiles();
  }
  clear_rect( 23, 0, 22, 79 );
  move (23, 0);
}


void end_game() {
  int i, j;
  int adj[4], wbs, wb[4], was, wa[4];
  long p1;
  FILE *fp;
  char c;

  clear_turn();
  printw( "*** GAME OVER\n" );
  fprintf(gcgptr,"#note Game over\n");
  fprintf(evalptr,"\n\n--- End of game\n");
  print_tiles();
  clear_prompt();
  printw( "  Here are the score adjustments: " );
  wbs = -1;
  /* wbs appears to be high score
     wb array marks who has it with 1 */
  for( i = 0; i < players; i++ )
    if( plr_scores[i] > wbs ) wbs = plr_scores[i];
  for( i = 0; i < players; i++ ) {
    if( plr_scores[i] == wbs ) wb[i] = 1;
    else wb[i] = 0;
  }
  /* go through each player's residual tiles 
     adj array is 1 when no tiles found?
     otherwise subtract each tile value from adj (weird)
     blanks are counted as 0?
  */
  for( i = 0; i < players; i++ ) {
    adj[i] = 1;
    for( j = 0; j < 7; j++ ) {
      c = plr_tiles[i][j];
      if( c != 0 ) {
	if( adj[i] == 1 ) adj[i] = 0;
	if( c != CH_BL )
	  adj[i] -= letters[c - 'A'].points;
      }
    }
  }
  /* now do nothing if adj=1 for player (no residuals) but if
     there is ONLY a blank, adj is still 0, not 1 */
  for( j = 0; j < players; j++ )
    if( adj[j] == 1 ) break;
  /* we break out when we get a player with adj=1 presumably the
     player with no tiles then subtract the negative value which
     means add positive! */
  /* we amend this to zero out the negative adjustment and
     double the positive adjustment as per NA rules */
  /* Looks like no adjustments if someone didn't go out? */
  if( j < players ) {
    adj[j] = 0;
    for( i = 0; i < players; i++ )
      if( i != j ) {
	adj[j] -= 2*adj[i];
	adj[i] = 0;
      }
  }
  /* Okay, it's right because the above j < players thing
     doesn't happen if no one goes out. Then residuals are
     simply subtracted. But this means more residual lines AND
     separate tile tracking. */
  for( i = 0; i < players; i++ ) {
    move( 19 + i, 36 );
    if( i == human_player ) addch( ' ' );
    printw( "%s: %d", you[i], adj[i] );
    plr_scores[i] += adj[i];
  }	

  /* and now with all adjustments made output the adjustment gcg lines */
  /* first, if we have someone who went out, one line only */
  if (j < players) {
    fprintf(gcgptr, ">%s: (", you[j]);
    for (i=0; i<players; i++)
      if (i != j)
	gcg_rack(i);
    fprintf(gcgptr, ") +%d %d\n",adj[j], plr_scores[j]); 
    /* no, no one went out */
  } else {
    for (i=0; i<players; i++) {
      fprintf(gcgptr, ">%s: ", you[i]);
      gcg_rack(i);
      fprintf(gcgptr, " (");
      /* Penalized letters. * should not be included? */
      gcg_rack(i);
      fprintf(gcgptr,") %d %d\n", adj[i], plr_scores[i]);
    }
  }
	
  /* now find new high score again */
  was = -1;
  for( i = 0; i < players; i++ )
    if( plr_scores[i] > was ) was = plr_scores[i];
  for( i = 0; i < players; i++ ) {
    if( plr_scores[i] == was ) wa[i] = 1;
    else wa[i] = 0;
  }
  press_return();
  print_tiles();
  clear_prompt();
  move( 20, 0 );
  was = 0;
  for( i = 0; i < players; i++ ) was += wa[i];
  wbs = 0;
  for( i = 0; i < players; i++ ) wbs += wb[i];
  if( was == 1 || wbs == 1 ) {
    if( was == 1 ) {
      for( i = 0; i < players; i++ )
	if( wa[i] == 1 ) break;
    } else {
      for( i = 0; i < players; i++ )
	if( wb[i] == 1 ) break;
    }
    printw( "  And the winner is:  %s!", you[i] );
    fprintf(gcgptr, "#note And the winner is:  %s!\n", you[i] );
  } else {	  
    printw( "  It was a tie between:  " );
    fprintf(gcgptr, "#note It was a tie between:  " );
    wbs = 1;
    for( i = 0; i < players; i++ )
      if( wa[i] == 1 ) {
	if( wbs < was - 1 ) printw( ", " );
	else {
	  if( was > 2 ) printw( ", and " );
	  else printw( " and " );
	  if( was > 2 ) fprintf(gcgptr, ", and " );
	  else fprintf(gcgptr, " and " );
	}
	printw( "%s", you[i] );
	fprintf(gcgptr, "%s", you[i] );
	wbs++;
      }
    printw( "!" );
    fprintf(gcgptr, "!\n");
  }
  refresh();
}

int main() {
  int plr;
  int i, plrstart;
  time_t now;
  char pname[20];
  int srstatus;

  open_log();
  set_up_window();
  print_version();

  read_words();
  dict_changed = 0;

  print_version();
  init_board();
  seed_random();
  print_board();
  print_info();
  print_first_tiles();
	
  sr_status = -1;
  on_turn = 0;
  game_done = 0;
  game_abort = 0;

  /* Initialize stats. */
  for (plr = 0; plr < 4; plr++) {
    turn[plr] = 0;
    cumpct[plr] = 0.0;
  }
  player_time = 0;
	
  /* Up-front restore will be done in get_num_players. */
  get_num_players();
  /* Set up handlers. Wait until after potential restore? */
  signal(SIGHUP, aborthandler);
  signal(SIGINT, aborthandler);
  signal(SIGQUIT, aborthandler);
  signal(SIGABRT, aborthandler);
  signal(SIGTERM, aborthandler);
  signal(SIGSEGV, aborthandler);
  /* Should there be more handlers?*/

  /* And here is the hard stuff on a restore. We have to set plr
     and jump in at the right place in the right way. */
  if (sr_status == -1) {
    /* GCG and EVALMOVE files set in restore. */
    /* Not a restore so open files if needed. */
    open_gcg();
    open_eval();
  }
  /* Now put in headers as needed. */
  time(&now);
  if (sr_status == -1)
    fprintf(logptr, "START       %s %s %s", gcgname, evalname, ctime(&now));	  
  fprintf(gcgptr,"#title SCRAMBLE %s %s", version, ctime(&now));
  for( i = 0; i < players; i++ ) {
    if (i == human_player) {
      strcpy(pname, "You ");
      strcat(pname, getenv("LOGNAME"));
    }  else {
      sprintf(pname, "%s %s Level %d", you[i], you[i], plr_skills[i]);
    }
    fprintf(gcgptr,"#player%d %s\n",i+1,pname);
  }
  fprintf(evalptr,"#title %s %s", version, ctime(&now));

  /* Show initial tiles remaining. */
  print_un_tiles();
  do {
    if (sr_status == 1) {
      plrstart = on_turn;
    } else {
      plrstart = 0;	    
    }
	  
    sr_status = 0;
    for( plr=plrstart; plr < players; plr++ ) {
      /* Here we'll calc next move (for potential save), which
	 is plr+1 mod players. */
      on_turn=(plr+1)%players;
      print_whos_move( plr );
      /* gcg scoring line */
      fprintf(gcgptr, ">%s: ", you[plr]);
      /* rack prior to move */
      gcg_rack(plr);

      if( plr == human_player ) {
	/* We call the computer move here to eval the player
	   situation. Just don't actually make the move! */
	computer_move (plr); 
	player_move();
      } else {
	computer_move( plr );
      }
      print_un_tiles();
      /* Stuff to do only if we haven't just restored
	 (normal case). */
      if (sr_status == 0) {
	if( game_abort != 0 ) break;
	print_passed_turns();
	if( game_done == players ) break;
	for( i = 0; i < 7; i++ )
	  if( plr_tiles[plr][i] != 0 ) break;
	if( i == 7 ) {
	  game_done = players;
	  break;
	}
      } else {
	/* If just restored need to break out of player move
	   loop. */
	break;
      }
    }
  } while( game_done < players && game_abort == 0 );
	
  if( game_abort == 0 ) {
    end_game();
  }
  else {
    fprintf(gcgptr,"#note Game aborted\n");
    fprintf(evalptr,"\n\n--- Game aborted\n");
    for (i=0;i<players; i++) {
      fprintf(gcgptr, "#rack%d ", i+1);
      gcg_rack(i);
      fprintf(gcgptr, "\n");
      /* We don't offer to save on game abort as the status
	 could be unsave-able. */
    }
  }
  /* Final exit with a borrowed line.*/
  thatsallfolks();
  /* Nevah go to Ewa. */
  return( 0 );
}
